from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter


class OutputWriter(AbstractOutputWriter):

    def write(self, msg: str):
        """Prints messages to STDOUT."""
        print(msg, flush=True)



from argparse import ArgumentParser
import boto3
from spotty.commands.abstract import AbstractCommand
from spotty.helpers.resources import is_valid_instance_type
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.helpers.spot_prices import get_spot_prices


class SpotPricesCommand(AbstractCommand):

    @staticmethod
    def get_name() -> str:
        return 'spot-prices'

    @staticmethod
    def get_description():
        return 'Get spot instance prices for particular instance type across all regions'

    @staticmethod
    def configure(parser: ArgumentParser):
        parser.add_argument('-i', '--instance-type', type=str, required=True, help='Instance type')
        parser.add_argument('-r', '--region', type=str, help='AWS region')

    def run(self, output: AbstractOutputWriter):
        # get all regions
        if not self._args.region:
            ec2 = boto3.client('ec2')
            res = ec2.describe_regions()
            regions = [row['RegionName'] for row in res['Regions']]
        else:
            regions = [self._args.region]

        instance_type = self._args.instance_type
        if not is_valid_instance_type(instance_type):
            raise ValueError('Instance type "%s" doesn\'t exist.' % instance_type)

        output.write('Getting spot instance prices for "%s"...\n' % instance_type)

        prices = []
        for region in regions:
            ec2 = boto3.client('ec2', region_name=region)
            res = get_spot_prices(ec2, instance_type)
            prices += [(price, zone) for zone, price in res.items()]

        # sort availability zones by price
        prices.sort(key=lambda x: x[0])

        if prices:
            output.write('Price  Zone')
            for price, zone in prices:
                output.write('%.04f %s' % (price, zone))
        else:
            output.write('Spot instances of this type are not available.')



import boto3
from spotty.commands.abstract_config import AbstractConfigCommand
from spotty.helpers.resources import wait_stack_status_changed
from spotty.helpers.validation import validate_instance_config
from spotty.project_resources.stack import StackResource
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter


class StopCommand(AbstractConfigCommand):

    @staticmethod
    def get_name() -> str:
        return 'stop'

    @staticmethod
    def get_description():
        return 'Terminate running instance and delete its stack'

    @staticmethod
    def _validate_config(config):
        return validate_instance_config(config)

    def run(self, output: AbstractOutputWriter):
        project_config = self._config['project']
        instance_config = self._config['instance']

        project_name = project_config['name']
        region = instance_config['region']

        cf = boto3.client('cloudformation', region_name=region)

        stack = StackResource(cf, project_name, region)

        # check that the stack exists
        if not stack.stack_exists():
            raise ValueError('Stack "%s" doesn\'t exists.' % stack.name)

        # get stack ID
        stack_info = stack.get_stack_info()
        stack_id = stack_info['StackId']

        # delete the stack
        stack.delete_stack()

        output.write('Waiting for the stack to be deleted...')

        resource_messages = [
            ('TerminateInstance', 'terminating the instance'),
            ('_', 'creating snapshots and deleting the volumes'),
        ]

        # wait for the deletion to be completed
        status, stack_info = wait_stack_status_changed(cf, stack_id=stack_id, waiting_status='DELETE_IN_PROGRESS',
                                                       resource_messages=resource_messages,
                                                       resource_success_status='DELETE_COMPLETE', output=output)

        if status == 'DELETE_COMPLETE':
            output.write('\n'
                         '--------------------\n'
                         'Stack was successfully deleted.\n'
                         '--------------------')
        else:
            raise ValueError('Stack "%s" was not deleted.\n'
                             'See CloudFormation and CloudWatch logs for details.' % stack.name)



import boto3
from spotty.commands.abstract_config import AbstractConfigCommand
from spotty.helpers.resources import wait_stack_status_changed, get_instance_ip_address, check_az_and_subnet
from spotty.helpers.spot_prices import get_current_spot_price
from spotty.helpers.sync import sync_project_with_s3
from spotty.helpers.validation import validate_instance_config
from spotty.project_resources.instance_profile_stack import create_or_update_instance_profile
from spotty.project_resources.stack import StackResource
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter


class StartCommand(AbstractConfigCommand):

    @staticmethod
    def get_name() -> str:
        return 'start'

    @staticmethod
    def get_description():
        return 'Run spot instance, sync the project and start the Docker container'

    @staticmethod
    def _validate_config(config):
        return validate_instance_config(config)

    def run(self, output: AbstractOutputWriter):
        project_config = self._config['project']
        instance_config = self._config['instance']

        region = instance_config['region']
        availability_zone = instance_config['availabilityZone']
        subnet_id = instance_config['subnetId']

        cf = boto3.client('cloudformation', region_name=region)
        ec2 = boto3.client('ec2', region_name=region)

        project_name = project_config['name']
        stack = StackResource(cf, project_name, region)

        # check if the stack already exists
        if stack.stack_exists():
            raise ValueError('Stack "%s" already exists.\n'
                             'Use "spotty stop" command to delete the stack.' % stack.name)

        # check availability zone and subnet
        check_az_and_subnet(ec2, availability_zone, subnet_id, region)

        # sync the project with S3
        output.write('Syncing the project with S3...')
        sync_filters = project_config['syncFilters']
        bucket_name = sync_project_with_s3(self._project_dir, project_name, region, sync_filters, output)

        # create or update instance profile
        instance_profile_arn = create_or_update_instance_profile(cf, output)

        # prepare CloudFormation template
        output.write('Preparing CloudFormation template...')

        instance_type = instance_config['instanceType']
        volumes = instance_config['volumes']
        ports = instance_config['ports']
        max_price = instance_config['maxPrice']
        on_demand = instance_config['onDemandInstance']
        docker_commands = instance_config['docker']['commands']

        template = stack.prepare_template(ec2, availability_zone, subnet_id, instance_type, volumes, ports, max_price,
                                          on_demand, docker_commands)

        # create stack
        ami_name = instance_config['amiName']
        root_volume_size = instance_config['rootVolumeSize']
        mount_dirs = [volume['directory'] for volume in volumes]
        docker_config = instance_config['docker']
        remote_project_dir = project_config['remoteDir']

        res = stack.create_stack(ec2, template, instance_profile_arn, instance_type, ami_name, root_volume_size,
                                 mount_dirs, bucket_name, remote_project_dir, project_name, self._project_dir,
                                 docker_config)

        output.write('Waiting for the stack to be created...')

        resource_messages = [
            ('SpotInstance', 'launching the instance'),
            ('DockerReadyWaitCondition', 'waiting for the Docker container to be ready'),
        ]

        # wait for the stack to be created
        status, info = wait_stack_status_changed(cf, stack_id=res['StackId'], waiting_status='CREATE_IN_PROGRESS',
                                                 resource_messages=resource_messages,
                                                 resource_success_status='CREATE_COMPLETE', output=output)

        if status == 'CREATE_COMPLETE':
            ip_address = get_instance_ip_address(ec2, stack.name)
            output.write('\n'
                         '--------------------\n'
                         'Instance is running.\n'
                         '\n'
                         'IP address: %s' % ip_address)

            if not on_demand:
                # get the current spot price
                availability_zone = [row['OutputValue'] for row in info['Outputs']
                                     if row['OutputKey'] == 'AvailabilityZone'][0]
                current_price = get_current_spot_price(ec2, instance_type, availability_zone)
                output.write('Current Spot price: $%.04f' % current_price)

            output.write('\n'
                         'Use "spotty ssh" command to connect to the Docker container.\n'
                         '--------------------')
        else:
            raise ValueError('Stack "%s" was not created.\n'
                             'Please, see CloudFormation and CloudWatch logs for the details.' % stack.name)



from argparse import ArgumentParser
import boto3
import subprocess
from spotty.commands.abstract_config import AbstractConfigCommand
from spotty.helpers.resources import get_instance_ip_address
from spotty.helpers.ssh import get_ssh_command
from spotty.helpers.validation import validate_instance_config
from spotty.project_resources.stack import StackResource
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter


class SshCommand(AbstractConfigCommand):

    @staticmethod
    def get_name() -> str:
        return 'ssh'

    @staticmethod
    def get_description():
        return 'Connect to the running Docker container or to the instance itself'

    @staticmethod
    def _validate_config(config):
        return validate_instance_config(config)

    @staticmethod
    def configure(parser: ArgumentParser):
        AbstractConfigCommand.configure(parser)
        parser.add_argument('--host-os', '-o', action='store_true', help='Connect to the host OS instead of the Docker '
                                                                         'container')
        parser.add_argument('--session-name', '-s', type=str, default=None, help='tmux session name')

    def run(self, output: AbstractOutputWriter):
        project_config = self._config['project']
        instance_config = self._config['instance']

        project_name = project_config['name']
        region = instance_config['region']
        local_ssh_port = instance_config['localSshPort']

        # get instance IP address
        stack = StackResource(None, project_name, region)
        ec2 = boto3.client('ec2', region_name=region)
        ip_address = get_instance_ip_address(ec2, stack.name)

        if self._args.host_os:
            # connect to the host OS
            session_name = self._args.session_name if self._args.session_name else 'spotty-ssh-host-os'
            remote_cmd = ['tmux', 'new', '-s', session_name, '-A']
        else:
            # connect to the container
            session_name = self._args.session_name if self._args.session_name else 'spotty-ssh-container'
            remote_cmd = ['tmux', 'new', '-s', session_name, '-A', 'sudo', '/scripts/container_bash.sh']

        remote_cmd = subprocess.list2cmdline(remote_cmd)

        # connect to the instance
        ssh_command = get_ssh_command(project_name, region, ip_address, remote_cmd, local_ssh_port)
        subprocess.call(ssh_command)



from spotty.project_resources.key_pair import KeyPairResource


def get_ssh_command(project_name: str, region: str, host: str, remote_cmd: str, local_ssh_port: None,
                    quiet: bool = False) -> list:
    ssh_port = 22
    user = 'ubuntu'

    if local_ssh_port:
        ssh_port = local_ssh_port
        host = '127.0.0.1'

    key_path = KeyPairResource(None, project_name, region).key_path

    ssh_command = ['ssh', '-p', str(ssh_port), '-i', key_path, '-o', 'StrictHostKeyChecking no', '-t']
    if quiet:
        ssh_command += ['-q']

    ssh_command += ['%s@%s' % (user, host), remote_cmd]

    return ssh_command



import os
import random
import string
import errno


def data_dir(path: str = ''):
    """Returns an absolute path to "data" directory.

    Args:
        path: A path which should be added to the "data" path.

    """
    res_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')
    if path:
        res_path = os.path.join(res_path, path)

    return res_path


def check_path(path):
    """Creates a directory if it doesn't exist."""
    if not os.path.exists(path):
        try:
            os.makedirs(path)
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise


def random_string(length: int, chars: str = string.ascii_lowercase + string.digits):
    return ''.join(random.choice(chars) for _ in range(length))



from argparse import ArgumentParser
from time import time
import boto3
from spotty.commands.abstract_config import AbstractConfigCommand
from spotty.helpers.validation import validate_logs_config
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter


class CleanLogsCommand(AbstractConfigCommand):

    @staticmethod
    def get_name() -> str:
        return 'clean-logs'

    @staticmethod
    def get_description():
        return 'Delete expired CloudFormation log groups with Spotty prefixes'

    @staticmethod
    def _validate_config(config):
        return validate_logs_config(config)

    @staticmethod
    def configure(parser: ArgumentParser):
        AbstractConfigCommand.configure(parser)
        parser.add_argument('--delete-all', '-a', action='store_true', help='Delete all Spotty log groups, '
                                                                            'not just expired ones')

    def run(self, output: AbstractOutputWriter):
        region = self._config['instance']['region']
        logs = boto3.client('logs', region_name=region)

        prefixes = ['spotty-', '/aws/lambda/spotty-']
        only_empty = not self._args.delete_all

        output.write('Deleting %s Spotty log groups...' % ('empty' if only_empty else 'all'))

        res = logs.describe_log_groups()
        self._delete_log_groups(logs, res['logGroups'], prefixes, only_empty, output)

        while 'nextToken' in res:
            res = logs.describe_log_groups(nextToken=res['nextToken'])
            self._delete_log_groups(logs, res['logGroups'], prefixes, only_empty, output)

        output.write('Done')

    @staticmethod
    def _delete_log_groups(logs, log_groups: list, prefixes: list, only_empty: bool, output: AbstractOutputWriter):
        for log_group in log_groups:
            for prefix in prefixes:
                if log_group['logGroupName'].startswith(prefix):
                    delete = True
                    if only_empty:
                        delete = False
                        days_passed = (int(time()) - log_group['creationTime'] // 1000) // 86400
                        if ('retentionInDays' in log_group) and (days_passed >= log_group['retentionInDays']):
                            delete = True

                    if delete:
                        output.write('[x] %s' % log_group['logGroupName'])
                        logs.delete_log_group(logGroupName=log_group['logGroupName'])
                    break



import json
import logging
import subprocess
from shutil import which


class AwsCommandError(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


class AwsCli(object):
    def __init__(self, profile: str = None, region: str = None):
        self._profile = profile
        self._region = region

    def s3_sync(self, from_path: str, to_path: str, delete=False, filters=None, capture_output=True) -> str:
        args = ['s3', 'sync', from_path, to_path]

        if delete:
            args.append('--delete')

        if filters:
            for sync_filter in filters:
                if ('exclude' in sync_filter and 'include' in sync_filter) \
                        or ('exclude' not in sync_filter and 'include' not in sync_filter):
                    raise ValueError('S3 sync filter has wrong format.')

                if 'exclude' in sync_filter:
                    for path in sync_filter['exclude']:
                        args += ['--exclude', path]

                if 'include' in sync_filter:
                    for path in sync_filter['include']:
                        args += ['--include', path]

        return self._run(args, False, capture_output=capture_output)

    def _run(self, args: list, json_format=True, capture_output=True):
        aws_cmd = 'aws'
        if which(aws_cmd) is None:
            raise ValueError('AWS CLI is not installed.')

        command_args = [aws_cmd]
        if self._profile:
            command_args += ['--profile', self._profile]

        if self._region:
            command_args += ['--region', self._region]

        if json_format:
            command_args += ['--output', 'json']

        command_args += args

        logging.debug('AWS command: ' + subprocess.list2cmdline(command_args))

        if capture_output:
            res = subprocess.run(command_args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            output = res.stdout.decode('utf-8')

            logging.debug('AWS command output: ' + output)

            if res.returncode:
                raise AwsCommandError(output)

            if json_format:
                output = json.loads(output)
        else:
            subprocess.run(command_args)
            output = None

        return output



import datetime


def get_spot_prices(ec2, instance_type: str):
    """Returns current Spot Instance prices for all availability zones for particular instance type and region.
    AWS region specified implicitly in the "ec2" object.
    """
    tomorrow_date = datetime.datetime.today() + datetime.timedelta(days=1)
    res = ec2.describe_spot_price_history(InstanceTypes=[instance_type],
                                          StartTime=tomorrow_date,
                                          ProductDescriptions=['Linux/UNIX'])

    prices_by_zone = {}
    for row in res['SpotPriceHistory']:
        prices_by_zone[row['AvailabilityZone']] = float(row['SpotPrice'])

    return prices_by_zone


def get_current_spot_price(ec2, instance_type, availability_zone=''):
    """Returns the current Spot price for an availability zone.
    If an availability zone is not specified, returns the minimum price for the region.
    """
    spot_prices = get_spot_prices(ec2, instance_type)
    if availability_zone:
        if availability_zone not in spot_prices:
            raise ValueError('Spot price for the "%s" availability zone not found.')

        current_price = spot_prices[availability_zone]
    else:
        current_price = min(spot_prices.values())

    return current_price



from time import sleep
from botocore.exceptions import EndpointConnectionError, WaiterError
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter


def get_snapshot(ec2, snapshot_name: str):
    """Returns a snapshot by its name."""
    res = ec2.describe_snapshots(Filters=[
        {'Name': 'tag:Name', 'Values': [snapshot_name]},
    ])

    if len(res['Snapshots']) > 1:
        raise ValueError('Several snapshots with Name=%s found.' % snapshot_name)

    snapshot = res['Snapshots'][0] if len(res['Snapshots']) else {}

    return snapshot


def get_volume(ec2, volume_name: str):
    """Returns the volume by its name."""
    res = ec2.describe_volumes(Filters=[
        {'Name': 'tag:Name', 'Values': [volume_name]},
        {'Name': 'status', 'Values': ['available']},
    ])

    if len(res['Volumes']) > 1:
        raise ValueError('Several volumes with Name=%s found.' % volume_name)

    volume = res['Volumes'][0] if len(res['Volumes']) else {}

    return volume


def get_instance_ip_address(ec2, stack_name):
    res = ec2.describe_instances(Filters=[
        {'Name': 'tag:aws:cloudformation:stack-name', 'Values': [stack_name]},
        {'Name': 'instance-state-name', 'Values': ['running']},
    ])

    if not len(res['Reservations']):
        raise ValueError('Instance is not running.\n'
                         'Use "spotty start" command to run an instance.')

    instance = res['Reservations'][0]['Instances'][0]
    ip_address = instance['PublicIpAddress'] if 'PublicIpAddress' in instance else instance['PrivateIpAddress']

    return ip_address


def get_default_subnet_ids(ec2):
    res = ec2.describe_subnets(Filters=[
        {'Name': 'defaultForAz', 'Values': ['true']},
    ])

    subnets_by_zones = {}
    for subnet in res['Subnets']:
        subnets_by_zones[subnet['AvailabilityZone']] = subnet['SubnetId']

    return subnets_by_zones


def get_ami(ec2, ami_name):
    res = ec2.describe_images(Owners=['self'], Filters=[
        {'Name': 'name', 'Values': [ami_name]},
    ])

    if len(res['Images']) > 1:
        raise ValueError('Several AMIs use the same name: "%s".' % ami_name)

    ami = {}
    if len(res['Images']):
        ami = res['Images'][0]

    return ami


def get_subnet(ec2, subnet_id):
    res = ec2.describe_subnets(Filters=[
        {'Name': 'subnet-id', 'Values': [subnet_id]},
    ])

    subnet = {}
    if len(res['Subnets']):
        subnet = res['Subnets'][0]

    return subnet


def stack_exists(cf, stack_name):
    res = True
    try:
        cf.get_waiter('stack_exists').wait(StackName=stack_name, WaiterConfig={'MaxAttempts': 1})
    except WaiterError:
        res = False

    return res


def wait_stack_status_changed(cf, stack_id, waiting_status, resource_messages, resource_success_status,
                              output: AbstractOutputWriter, delay=5):
    current_status = waiting_status
    stack = None

    resource_messages = iter(resource_messages) if resource_messages else None
    resource_name = None

    while current_status == waiting_status:
        sleep(delay)

        # display resource creation progress
        if resource_messages:
            try:
                stack_resources = cf.list_stack_resources(StackName=stack_id)
            except EndpointConnectionError:
                output.write('Connection problem')
                continue

            resource_statuses = dict([(row['LogicalResourceId'], row['ResourceStatus'])
                                      for row in stack_resources['StackResourceSummaries']])

            while (resource_name is None) or (resource_name and
                                              resource_statuses.get(resource_name, '') == resource_success_status):
                (resource_name, resource_msg) = next(resource_messages, (False, False))
                if resource_name:
                    output.write('  - %s...' % resource_msg)

        # get the latest status of the stack
        try:
            res = cf.describe_stacks(StackName=stack_id)
        except EndpointConnectionError:
            output.write('Connection problem')
            continue

        stack = res['Stacks'][0]
        current_status = stack['StackStatus']

    return current_status, stack


def check_az_and_subnet(ec2, availability_zone, subnet_id, region):
    # get all availability zones for the region
    zones = ec2.describe_availability_zones()
    zone_names = [zone['ZoneName'] for zone in zones['AvailabilityZones']]

    # check availability zone
    if availability_zone and availability_zone not in zone_names:
        raise ValueError('Availability zone "%s" doesn\'t exist in the "%s" region.'
                         % (availability_zone, region))

    if availability_zone:
        if subnet_id:
            subnet = get_subnet(ec2, subnet_id)
            if not subnet:
                raise ValueError('Subnet "%s" not found.' % subnet_id)

            if subnet['AvailabilityZone'] != availability_zone:
                raise ValueError('Availability zone of the subnet doesn\'t match the specified availability zone')
        else:
            default_subnets = get_default_subnet_ids(ec2)
            if availability_zone not in default_subnets:
                raise ValueError('Default subnet for the "%s" availability zone not found.\n'
                                 'Use the "subnetId" parameter to specify a subnet for this availability zone.'
                                 % availability_zone)
    else:
        if subnet_id:
            raise ValueError('An availability zone should be specified if a custom subnet is used.')
        else:
            default_subnets = get_default_subnet_ids(ec2)
            zones_wo_subnet = [zone_name for zone_name in zone_names if zone_name not in default_subnets]
            if zones_wo_subnet:
                raise ValueError('Default subnets for the following availability zones were not found: %s.\n'
                                 'Use "subnetId" and "availabilityZone" parameters or create missing default subnets.'
                                 % ', '.join(zones_wo_subnet))


def is_gpu_instance(instance_type: str):
    return instance_type in [
        'p2.xlarge', 'p2.8xlarge', 'p2.16xlarge',
        'p3.2xlarge', 'p3.8xlarge', 'p3.16xlarge',
        'g3s.xlarge', 'g3.4xlarge', 'g3.8xlarge', 'g3.16xlarge',
    ]


def is_valid_instance_type(instance_type: str):
    return instance_type in [
        't1.micro',
        't2.nano', 't2.micro', 't2.small', 't2.medium', 't2.large', 't2.xlarge', 't2.2xlarge',
        'm1.small', 'm1.medium', 'm1.large', 'm1.xlarge',
        'm3.medium', 'm3.large', 'm3.xlarge', 'm3.2xlarge',
        'm4.large', 'm4.xlarge', 'm4.2xlarge', 'm4.4xlarge', 'm4.10xlarge', 'm4.16xlarge',
        'm2.xlarge', 'm2.2xlarge', 'm2.4xlarge',
        'cr1.8xlarge',
        'r3.large', 'r3.xlarge', 'r3.2xlarge', 'r3.4xlarge', 'r3.8xlarge',
        'r4.large', 'r4.xlarge', 'r4.2xlarge', 'r4.4xlarge', 'r4.8xlarge', 'r4.16xlarge',
        'x1.16xlarge', 'x1.32xlarge',
        'x1e.xlarge', 'x1e.2xlarge', 'x1e.4xlarge', 'x1e.8xlarge', 'x1e.16xlarge', 'x1e.32xlarge',
        'i2.xlarge', 'i2.2xlarge', 'i2.4xlarge', 'i2.8xlarge',
        'i3.large', 'i3.xlarge', 'i3.2xlarge', 'i3.4xlarge', 'i3.8xlarge', 'i3.16xlarge', 'i3.metal',
        'hi1.4xlarge',
        'hs1.8xlarge',
        'c1.medium', 'c1.xlarge',
        'c3.large', 'c3.xlarge', 'c3.2xlarge', 'c3.4xlarge', 'c3.8xlarge',
        'c4.large', 'c4.xlarge', 'c4.2xlarge', 'c4.4xlarge', 'c4.8xlarge',
        'c5.large', 'c5.xlarge', 'c5.2xlarge', 'c5.4xlarge', 'c5.9xlarge', 'c5.18xlarge',
        'c5d.large', 'c5d.xlarge', 'c5d.2xlarge', 'c5d.4xlarge', 'c5d.9xlarge', 'c5d.18xlarge',
        'cc1.4xlarge',
        'cc2.8xlarge',
        'g2.2xlarge', 'g2.8xlarge',
        'g3s.xlarge', 'g3.4xlarge', 'g3.8xlarge', 'g3.16xlarge',
        'cg1.4xlarge',
        'p2.xlarge', 'p2.8xlarge', 'p2.16xlarge',
        'p3.2xlarge', 'p3.8xlarge', 'p3.16xlarge',
        'd2.xlarge', 'd2.2xlarge', 'd2.4xlarge', 'd2.8xlarge',
        'f1.2xlarge', 'f1.16xlarge',
        'm5.large', 'm5.xlarge', 'm5.2xlarge', 'm5.4xlarge', 'm5.12xlarge', 'm5.24xlarge',
        'm5d.large', 'm5d.xlarge', 'm5d.2xlarge', 'm5d.4xlarge', 'm5d.12xlarge', 'm5d.24xlarge',
        'h1.2xlarge', 'h1.4xlarge', 'h1.8xlarge', 'h1.16xlarge',
    ]



import boto3
from spotty.commands.abstract_config import AbstractConfigCommand
from spotty.helpers.resources import wait_stack_status_changed, get_ami
from spotty.helpers.validation import validate_ami_config
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter


class DeleteAmiCommand(AbstractConfigCommand):

    @staticmethod
    def get_name() -> str:
        return 'delete-ami'

    @staticmethod
    def get_description():
        return 'Delete AMI with NVIDIA Docker'

    @staticmethod
    def _validate_config(config):
        return validate_ami_config(config)

    def run(self, output: AbstractOutputWriter):
        region = self._config['instance']['region']
        cf = boto3.client('cloudformation', region_name=region)
        ec2 = boto3.client('ec2', region_name=region)

        # get image info
        ami_name = self._config['instance']['amiName']
        ami_info = get_ami(ec2, ami_name)

        # check that the image exists
        if not ami_info:
            raise ValueError('AMI with name "%s" not found.' % ami_name)

        # get stack ID for the image
        tag_values = [tag['Value'] for tag in ami_info['Tags'] if tag['Key'] == 'spotty:stack-id']
        if not len(tag_values):
            raise ValueError('AMI wasn\'t created by Spotty')

        # ask user to confirm the deletion
        ami_id = ami_info['ImageId']
        confirm = input('AMI "%s" (ID=%s) will be deleted.\n'
                        'Type "y" to confirm: '
                        % (ami_name, ami_id))
        if confirm != 'y':
            output.write('You didn\'t confirm the operation.')
            return

        # delete the image
        stack_id = tag_values[0]
        cf.delete_stack(StackName=stack_id)

        output.write('Waiting for the AMI to be deleted...')

        # wait for the deletion to be completed
        status, stack = wait_stack_status_changed(cf, stack_id=stack_id, waiting_status='DELETE_IN_PROGRESS',
                                                  resource_messages=[],
                                                  resource_success_status='DELETE_COMPLETE', output=output)

        if status == 'DELETE_COMPLETE':
            output.write('\n'
                         '--------------------\n'
                         'AMI was successfully deleted.\n'
                         '--------------------')
        else:
            raise ValueError('Stack "%s" not deleted.\n'
                             'See CloudFormation and CloudWatch logs for details.' % stack_id)



import yaml
from cfn_tools import CfnYamlLoader, CfnYamlDumper
from spotty.utils import data_dir, random_string


class AmiStackResource(object):

    def __init__(self, cf):
        self._cf = cf

    def prepare_template(self, availability_zone: str, subnet_id: str, on_demand: bool, key_name: str):
        """Prepares CloudFormation template to run a Spot Instance."""

        # read and update CF template
        with open(data_dir('create_ami.yaml')) as f:
            template = yaml.load(f, Loader=CfnYamlLoader)

        # remove key parameter if key is not provided
        if not key_name:
            del template['Parameters']['KeyName']
            del template['Resources']['SpotInstanceLaunchTemplate']['Properties']['LaunchTemplateData']['KeyName']

        # set availability zone
        if availability_zone:
            template['Resources']['SpotInstanceLaunchTemplate']['Properties']['LaunchTemplateData']['Placement'] = {
                'AvailabilityZone': availability_zone,
            }

        # set subnet
        if subnet_id:
            template['Resources']['SpotInstanceLaunchTemplate']['Properties']['LaunchTemplateData']['NetworkInterfaces'] = [{
                'SubnetId': subnet_id,
                'DeviceIndex': 0,
            }]

        # run on-demand instance
        if on_demand:
            del template['Resources']['SpotInstanceLaunchTemplate']['Properties']['LaunchTemplateData']['InstanceMarketOptions']

        return yaml.dump(template, Dumper=CfnYamlDumper)

    def create_stack(self, template: str, instance_type: str, ami_name: str, key_name: str):
        """Runs CloudFormation template."""
        params = {
            'InstanceType': instance_type,
            'ImageName': ami_name,
        }

        if key_name:
            params['KeyName'] = key_name

        stack_name = 'spotty-nvidia-docker-ami-%s' % random_string(8)
        res = self._cf.create_stack(
            StackName=stack_name,
            TemplateBody=template,
            Parameters=[{'ParameterKey': key, 'ParameterValue': value} for key, value in params.items()],
            Capabilities=['CAPABILITY_IAM'],
            OnFailure='DELETE',
        )

        return res, stack_name



import os


def get_spotty_config_dir():
    """Spotty configuration directory."""
    path = os.path.join(os.path.expanduser('~'), '.spotty')
    if not os.path.isdir(path):
        os.makedirs(path, mode=0o755, exist_ok=True)

    return path


def get_spotty_keys_dir():
    """"Spotty keys directory."""
    path = os.path.join(get_spotty_config_dir(), 'keys')
    if not os.path.isdir(path):
        os.makedirs(path, mode=0o755, exist_ok=True)

    return path



import os
from botocore.exceptions import ClientError
from spotty.configuration import get_spotty_keys_dir


class KeyPairResource(object):

    def __init__(self, ec2, project_name: str, region: str):
        self._ec2 = ec2
        self._region = region
        self._key_name = 'spotty-%s-%s' % (project_name, region)

    @property
    def key_path(self):
        return os.path.join(get_spotty_keys_dir(), self._key_name)

    def create_key(self):
        key_path = self.key_path
        key_file_exists = os.path.isfile(key_path)
        aws_key_exists = self._key_exists()

        if not aws_key_exists or not key_file_exists:
            # remove key from AWS (key file not found)
            if aws_key_exists:
                self._ec2.delete_key_pair(KeyName=self._key_name)

            # create new key
            res = self._ec2.create_key_pair(KeyName=self._key_name)

            # save the key to spotty directory
            with open(key_path, 'w') as f:
                f.write(res['KeyMaterial'])

            os.chmod(key_path, 0o600)

        return self._key_name

    def delete_key(self):
        if self._key_exists():
            self._ec2.delete_key_pair(KeyName=self._key_name)

        key_path = self.key_path
        if os.path.isfile(key_path):
            os.unlink(key_path)

    def _key_exists(self):
        key_exists = True
        try:
            self._ec2.describe_key_pairs(KeyNames=[self._key_name])
        except ClientError:
            key_exists = False

        return key_exists
